
/*********************************************************************/
/*                   BARE-BONES NETFLOW COLLECTOR                    */
/*                                                                   */
/*                Written by Martin van den Nieuwelaar               */
/*                   Contact: martin@gadgets.co.nz                   */
/*                         www.gadgets.co.nz                         */
/*                                                                   */
/* Gathers and displays Netflow Exports in real-time                 */
/* Intended to be used as a tool for debugging Netflow problems      */
/*                                                                   */
/* This source is free.  Do what you like with it.                   */
/*                                                                   */
/* Please direct comments and requests for features to my E-Mail     */
/* address.  If I'm not too busy and they're reasonable, I'll be     */
/* happy to add them.                                                */
/*                                                                   */
/*********************************************************************/

/*********************************************************************/
/* Known bugs:                                                       */
/*                                                                   */
/* getprotobynumber() appears to leak memory, at least under Linux.  */
/* No known workaround.                                              */
/*                                                                   */
/* V1 flow collection untested.  If it works, please let me know.    */
/* V7 flow collection untested.  If it works, please let me know.    */
/* V8 flow collection untested.  If it works, please let me know.    */
/* Only the first 5 types of V8 aggregation are presently supported. */
/*                                                                   */
/*********************************************************************/

/*********************************************************************/
/* 1.00  - Initial release                                           */
/* 1.01  - Checks for failure to bind to port and bails out          */
/* 1.02  - Fixes to timestamp display                                */
/*         Optional command line parameter for port to listen on     */
/* 1.03  - Fixes to format of printed values                         */
/* 1.04a - Added V7 flow support                                     */
/* 1.05a - Added V1 and V8 flow support.                             */
/*         pad1 was of the wrong size for V5 exports.  Fixed.        */
/* 1.05b - Removed message regarding supported versions.             */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "netflow.h"

#define VERSION "V1.05b"

#define DEFAULT_SERV_PORT 2055

/* The largest chunk of data we will read in from a socket */
#define MAX_BUFF_SIZE 16384


void usage(char *name)
{
  printf("usage: %s [-h][-p <port_number>]\n\n", name);

  printf("-h                  Help of command usage\n");
  printf("-p <port_number>    UDP port to listen on\n");
}


void display_v1_header(char *buff)
{
  struct netflow_export_v1_header *header =
    (struct netflow_export_v1_header *)buff;
  time_t uptime = ntohl(header->unix_secs);
  time_t boot_time = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000;

  printf("Version number  = %hu\n", ntohs(header->version));
  printf("Number of flows = %hu\n", ntohs(header->count));

  printf("SysUptime       = %lums or booted at %s",
	 (unsigned long)ntohl(header->SysUptime),
	 ctime(&boot_time));

  printf("unix_secs       = %lus or %s",
	 (unsigned long)ntohl(header->unix_secs),
	 ctime(&uptime));

  printf("unix_nsecs      = %luns\n",
	 (unsigned long)ntohl(header->unix_nsecs));
}


void display_v1_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v1_header *header =
    (struct netflow_export_v1_header *)header_buff;
  struct netflow_export_v1_flow *flow =
    (struct netflow_export_v1_flow *)flow_buff;
  struct protoent *protocol;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("Source IP address      = %s\n", inet_ntoa(flow->srcaddr));
  printf("Destination IP address = %s\n", inet_ntoa(flow->dstaddr));
  printf("Next hop address       = %s\n", inet_ntoa(flow->nexthop));

  printf("input (SNMP ifIndex)   = %hu\n", ntohs(flow->input));
  printf("output (SNMP ifIndex)  = %hu\n", ntohs(flow->output));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("srcport                = %hu\n", ntohs(flow->srcport));
  printf("dstport                = %hu\n", ntohs(flow->dstport));
  printf("pad1                   = %hu\n", ntohs(flow->pad1));

  printf("prot                   = %u ", flow->prot);

  /* getprotobynumber appears to cause a memory leak under Linux! */
  /* Cannot find a way around this at present. */

  if ((protocol = getprotobynumber(flow->prot)) != NULL)
    if (protocol->p_name != NULL)
      printf("(%s)", protocol->p_name);

  printf("\n");

  printf("tos                    = %u\n", flow->tos);
  printf("tcp_flags              = %u\n", flow->tcp_flags);
  printf("pad2                   = %u\n", flow->pad2);
  printf("pad3                   = %u\n", flow->pad3);
  printf("pad4                   = %u\n", flow->pad4);

  printf("reserved               = %lu\n",
    (unsigned long)ntohl(flow->reserved));
}


void display_v5_header(char *buff)
{
  struct netflow_export_v5_header *header =
    (struct netflow_export_v5_header *)buff;
  time_t uptime = ntohl(header->unix_secs);
  time_t boot_time = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000;

  printf("Version number  = %hu\n", ntohs(header->version));
  printf("Number of flows = %hu\n", ntohs(header->count));

  printf("SysUptime       = %lums or booted at %s",
	 (unsigned long)ntohl(header->SysUptime),
	 ctime(&boot_time));

  printf("unix_secs       = %lus or %s",
	 (unsigned long)ntohl(header->unix_secs),
	 ctime(&uptime));

  printf("unix_nsecs      = %luns\n",
	 (unsigned long)ntohl(header->unix_nsecs));

  printf("flow_sequence   = %lu\n",
	 (unsigned long)ntohl(header->flow_sequence));

  printf("engine_type     = %u\n", header->engine_type);
  printf("engine_id       = %u\n", header->engine_id);
  printf("reserved        = %hu\n", ntohs(header->reserved));
}


void display_v5_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v5_header *header =
    (struct netflow_export_v5_header *)header_buff;
  struct netflow_export_v5_flow *flow =
    (struct netflow_export_v5_flow *)flow_buff;
  struct protoent *protocol;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("Source IP address      = %s\n", inet_ntoa(flow->srcaddr));
  printf("Destination IP address = %s\n", inet_ntoa(flow->dstaddr));
  printf("Next hop address       = %s\n", inet_ntoa(flow->nexthop));

  printf("input (SNMP ifIndex)   = %hu\n", ntohs(flow->input));
  printf("output (SNMP ifIndex)  = %hu\n", ntohs(flow->output));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("srcport                = %hu\n", ntohs(flow->srcport));
  printf("dstport                = %hu\n", ntohs(flow->dstport));
  printf("pad1                   = %u\n", flow->pad1);
  printf("tcp_flags              = %u\n", flow->tcp_flags);

  printf("prot                   = %u ", flow->prot);

  /* getprotobynumber appears to cause a memory leak under Linux! */
  /* Cannot find a way around this at present. */

  if ((protocol = getprotobynumber(flow->prot)) != NULL)
    if (protocol->p_name != NULL)
      printf("(%s)", protocol->p_name);

  printf("\n");

  printf("tos                    = %u\n", flow->tos);
  printf("src_as                 = %hu\n", ntohs(flow->src_as));
  printf("dst_as                 = %hu\n", ntohs(flow->dst_as));
  printf("src_mask               = %u\n", flow->src_mask);
  printf("dst_mask               = %u\n", flow->dst_mask);
  printf("pad2                   = %hu\n", ntohs(flow->pad2));
}


void display_v7_header(char *buff)
{
  struct netflow_export_v7_header *header =
    (struct netflow_export_v7_header *)buff;
  time_t uptime = ntohl(header->unix_secs);
  time_t boot_time = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000;

  printf("Version number  = %hu\n", ntohs(header->version));
  printf("Number of flows = %hu\n", ntohs(header->count));

  printf("SysUptime       = %lums or booted at %s",
	 (unsigned long)ntohl(header->SysUptime),
	 ctime(&boot_time));

  printf("unix_secs       = %lus or %s",
	 (unsigned long)ntohl(header->unix_secs),
	 ctime(&uptime));

  printf("unix_nsecs      = %luns\n",
	 (unsigned long)ntohl(header->unix_nsecs));

  printf("flow_sequence   = %lu\n",
	 (unsigned long)ntohl(header->flow_sequence));

  printf("reserved        = %lu\n",
	 (unsigned long)ntohl(header->reserved));
}


void display_v7_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v7_header *header =
    (struct netflow_export_v7_header *)header_buff;
  struct netflow_export_v7_flow *flow =
    (struct netflow_export_v7_flow *)flow_buff;
  struct protoent *protocol;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("Source IP address      = %s\n", inet_ntoa(flow->srcaddr));
  printf("Destination IP address = %s\n", inet_ntoa(flow->dstaddr));
  printf("Next hop address       = %s\n", inet_ntoa(flow->nexthop));

  printf("input (SNMP ifIndex)   = %hu\n", ntohs(flow->input));
  printf("output (SNMP ifIndex)  = %hu\n", ntohs(flow->output));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("srcport                = %hu\n", ntohs(flow->srcport));
  printf("dstport                = %hu\n", ntohs(flow->dstport));
  printf("field flags            = %u\n", flow->field_flags);
  printf("tcp_flags              = %u\n", flow->tcp_flags);

  printf("prot                   = %u ", flow->prot);

  /* getprotobynumber appears to cause a memory leak under Linux! */
  /* Cannot find a way around this at present. */

  if ((protocol = getprotobynumber(flow->prot)) != NULL)
    if (protocol->p_name != NULL)
      printf("(%s)", protocol->p_name);

  printf("\n");

  printf("tos                    = %u\n", flow->tos);
  printf("src_as                 = %hu\n", ntohs(flow->src_as));
  printf("dst_as                 = %hu\n", ntohs(flow->dst_as));
  printf("src_mask               = %u\n", flow->src_mask);
  printf("dst_mask               = %u\n", flow->dst_mask);

  printf("flow flags             = %hu\n", ntohs(flow->flow_flags));
  printf("router_sc              = %s\n", inet_ntoa(flow->router_sc));
}


void display_v8_header(char *buff)
{
  struct netflow_export_v8_header *header =
    (struct netflow_export_v8_header *)buff;
  time_t uptime = ntohl(header->unix_secs);
  time_t boot_time = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000;

  printf("Version number  = %hu\n", ntohs(header->version));
  printf("Number of flows = %hu\n", ntohs(header->count));

  printf("SysUptime       = %lums or booted at %s",
	 (unsigned long)ntohl(header->SysUptime),
	 ctime(&boot_time));

  printf("unix_secs       = %lus or %s",
	 (unsigned long)ntohl(header->unix_secs),
	 ctime(&uptime));

  printf("unix_nsecs      = %luns\n",
	 (unsigned long)ntohl(header->unix_nsecs));

  printf("flow_sequence   = %lu\n",
	 (unsigned long)ntohl(header->flow_sequence));

  printf("engine_type     = %u\n", header->engine_type);
  printf("engine_id       = %u\n", header->engine_id);
  printf("aggregation     = %u\n", header->aggregation);
  printf("agg_version     = %u\n", header->agg_version);
  printf("sampling_intval = %hu\n", ntohs(header->sampling_interval));
  printf("reserved        = %hu\n", ntohs(header->reserved));
}


void display_v8_as_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v8_header *header =
    (struct netflow_export_v8_header *)header_buff;
  struct netflow_export_v8_as_flow *flow =
    (struct netflow_export_v8_as_flow *)flow_buff;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("flows                  = %lu\n", (unsigned long)ntohl(flow->flows));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("src_as                 = %hu\n", ntohs(flow->src_as));
  printf("dst_as                 = %hu\n", ntohs(flow->dst_as));
  printf("input (SNMP ifIndex)   = %hu\n", ntohs(flow->input));
  printf("output (SNMP ifIndex)  = %hu\n", ntohs(flow->output));
}


void display_v8_protoport_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v8_header *header =
    (struct netflow_export_v8_header *)header_buff;
  struct netflow_export_v8_protoport_flow *flow =
    (struct netflow_export_v8_protoport_flow *)flow_buff;
  struct protoent *protocol;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("flows                  = %lu\n", (unsigned long)ntohl(flow->flows));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("prot                   = %u ", flow->prot);

  /* getprotobynumber appears to cause a memory leak under Linux! */
  /* Cannot find a way around this at present. */

  if ((protocol = getprotobynumber(flow->prot)) != NULL)
    if (protocol->p_name != NULL)
      printf("(%s)", protocol->p_name);

  printf("\n");

  printf("pad                    = %u\n", flow->pad);

  printf("reserved               = %hu\n", ntohs(flow->reserved));
  printf("srcport                = %hu\n", ntohs(flow->srcport));
  printf("dstport                = %hu\n", ntohs(flow->dstport));
}


void display_v8_dstprefix_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v8_header *header =
    (struct netflow_export_v8_header *)header_buff;
  struct netflow_export_v8_dstprefix_flow *flow =
    (struct netflow_export_v8_dstprefix_flow *)flow_buff;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("flows                  = %lu\n", (unsigned long)ntohl(flow->flows));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("dst_prefix             = %s\n", inet_ntoa(flow->dst_prefix));

  printf("dst_mask               = %u\n", flow->dst_mask);
  printf("pad                    = %u\n", flow->pad);

  printf("dst_as                 = %hu\n", ntohs(flow->dst_as));
  printf("output (SNMP ifIndex)  = %hu\n", ntohs(flow->output));
  printf("reserved               = %hu\n", ntohs(flow->reserved));
}


void display_v8_srcprefix_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v8_header *header =
    (struct netflow_export_v8_header *)header_buff;
  struct netflow_export_v8_srcprefix_flow *flow =
    (struct netflow_export_v8_srcprefix_flow *)flow_buff;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("flows                  = %lu\n", (unsigned long)ntohl(flow->flows));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("src_prefix             = %s\n", inet_ntoa(flow->src_prefix));

  printf("src_mask               = %u\n", flow->src_mask);
  printf("pad                    = %u\n", flow->pad);

  printf("src_as                 = %hu\n", ntohs(flow->src_as));
  printf("input (SNMP ifIndex)   = %hu\n", ntohs(flow->input));
  printf("reserved               = %hu\n", ntohs(flow->reserved));
}


void display_v8_prefix_flow(char *header_buff, char *flow_buff)
{
  struct netflow_export_v8_header *header =
    (struct netflow_export_v8_header *)header_buff;
  struct netflow_export_v8_prefix_flow *flow =
    (struct netflow_export_v8_prefix_flow *)flow_buff;

  time_t flow_start = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->First)/1000;
  time_t flow_finish = ntohl(header->unix_secs)-ntohl(header->SysUptime)/1000+ntohl(flow->Last)/1000;

  printf("flows                  = %lu\n", (unsigned long)ntohl(flow->flows));
  printf("dPkts                  = %lu\n", (unsigned long)ntohl(flow->dPkts));

  printf("dOctets                = %lu bytes (layer 3 payload)\n",
	 (unsigned long)ntohl(flow->dOctets));

  printf("First                  = %lums or %s",
	 (unsigned long)ntohl(flow->First),
	 ctime(&flow_start));

  printf("Last                   = %lums or %s",
	 (unsigned long)ntohl(flow->Last),
	 ctime(&flow_finish));

  printf("src_prefix             = %s\n", inet_ntoa(flow->src_prefix));
  printf("dst_prefix             = %s\n", inet_ntoa(flow->dst_prefix));

  printf("src_mask               = %u\n", flow->src_mask);
  printf("dst_mask               = %u\n", flow->dst_mask);

  printf("reserved               = %hu\n", ntohs(flow->reserved));
  printf("src_as                 = %hu\n", ntohs(flow->src_as));
  printf("dst_as                 = %hu\n", ntohs(flow->dst_as));
  printf("input (SNMP ifIndex)   = %hu\n", ntohs(flow->input));
  printf("output (SNMP ifIndex)  = %hu\n", ntohs(flow->output));
}


int number_of_flows(char *buff)
     /* Routine to return to number of flows specified by the passed */
     /* flow header.  All flow headers, regardless of version contain */
     /* the number of flows in the same two bytes. */
{
  /* Doesn't matter what header type we specify here.  */
  /* For our purposes they're all the same */

  struct netflow_export_v5_header *header =
    (struct netflow_export_v5_header *)buff;

  return ntohs(header->count);
}


int version(char *buff)
     /* Routine to return the version number of the passed flow header. */
     /* All flow headers, regardless of version contain the version */
     /* number in the first two bytes. */
{
  /* Doesn't matter what header type we specify here.  */
  /* For our purposes they're all the same */

  struct netflow_export_v5_header *header =
    (struct netflow_export_v5_header *)buff;

  return ntohs(header->version);
}


int aggregation_type(char *buff)
     /* Routine to return the aggregation type being used in a version 8 */
     /* netflow export */
{
  struct netflow_export_v8_header *header =
    (struct netflow_export_v8_header *)buff;

  return ntohs(header->agg_version);
}


void display_v1(char *buff, int num_flows)
{
  int i;

  display_v1_header(buff);

  if (num_flows > NETFLOW_V1_MAX_FLOWS)
  {
    printf("Malformed header (more than %d flows in export).  Skipping...\n",
	   NETFLOW_V1_MAX_FLOWS);
    return;
  }

  for (i = 0; i < num_flows; i++)
  {
    printf("\n");
    display_v1_flow(buff,
		    buff+NETFLOW_V1_HEADER_LENGTH+i*NETFLOW_V1_FLOW_LENGTH);
  }

  printf("--\n");
}


void display_v5(char *buff, int num_flows)
{
  int i;

  display_v5_header(buff);

  if (num_flows > NETFLOW_V5_MAX_FLOWS)
  {
    printf("Malformed header (more than %d flows in export).  Skipping...\n",
	   NETFLOW_V5_MAX_FLOWS);
    return;
  }

  for (i = 0; i < num_flows; i++)
  {
    printf("\n");
    display_v5_flow(buff,
		    buff+NETFLOW_V5_HEADER_LENGTH+i*NETFLOW_V5_FLOW_LENGTH);
  }

  printf("--\n");
}


void display_v7(char *buff, int num_flows)
{
  int i;

  display_v7_header(buff);

  if (num_flows > NETFLOW_V7_MAX_FLOWS)
  {
    printf("Malformed header (more than %d flows in export).  Skipping...\n",
	   NETFLOW_V7_MAX_FLOWS);
    return;
  }

  for (i = 0; i < num_flows; i++)
  {
    printf("\n");
    display_v7_flow(buff,
		    buff+NETFLOW_V7_HEADER_LENGTH+i*NETFLOW_V7_FLOW_LENGTH);
  }

  printf("--\n");
}


void display_v8(char *buff, int num_flows)
{
  int i, type;

  display_v8_header(buff);

  if (num_flows > NETFLOW_V8_MAX_FLOWS)
  {
    printf("Malformed header (more than %d flows in export).  Skipping...\n",
	   NETFLOW_V8_MAX_FLOWS);
    return;
  }

  type = aggregation_type(buff);

  for (i = 0; i < num_flows; i++)
  {
    printf("\n");

    /* There is a different display method for each type */

    switch (type)
    {
    case V8_AS_FLOW_AGGREGATION:
      display_v8_as_flow(buff, buff+NETFLOW_V8_HEADER_LENGTH+i*NETFLOW_V8_AS_FLOW_LENGTH);
      break;

    case V8_PROTOPORT_FLOW_AGGREGATION:
      display_v8_protoport_flow(buff, buff+NETFLOW_V8_HEADER_LENGTH+i*NETFLOW_V8_PROTOPORT_FLOW_LENGTH);
      break;

    case V8_SRCPREFIX_FLOW_AGGREGATION:
      display_v8_srcprefix_flow(buff, buff+NETFLOW_V8_HEADER_LENGTH+i*NETFLOW_V8_SRCPREFIX_FLOW_LENGTH);
      break;

    case V8_DSTPREFIX_FLOW_AGGREGATION:
      display_v8_dstprefix_flow(buff, buff+NETFLOW_V8_HEADER_LENGTH+i*NETFLOW_V8_DSTPREFIX_FLOW_LENGTH);
      break;

    case V8_PREFIX_FLOW_AGGREGATION:
      display_v8_prefix_flow(buff, buff+NETFLOW_V8_HEADER_LENGTH+i*NETFLOW_V8_PREFIX_FLOW_LENGTH);
      break;

    default:
      printf("Unsupported aggregation type %d.  Skipping...\n", type);
    }
  }

  printf("--\n");
}


void display_nfe(char *buff, int buff_len)
{
  int nof; /* Number of flows in this transmission */
  int v;         /* Version of this netflow export */

  printf("Total export size %d bytes\n\n", buff_len);

  if (buff_len < NETFLOW_MINIMUM_HEADER_LENGTH)
  {
    printf("Malformed header (less than %d bytes long).  Skipping...\n",
	   NETFLOW_MINIMUM_HEADER_LENGTH);
    return;
  }

  nof = number_of_flows(buff);

  switch (v = version(buff))
    {
    case 1:
      display_v1(buff, nof);
      break;

    case 5:
      display_v5(buff, nof);
      break;

    case 7:
      display_v7(buff, nof);
      break;

    case 8:
      display_v8(buff, nof);
      break;

    default:
      printf("Unsupported netflow export type V%d.  Skipping...\n", v);
    }
}


int main(int argc, char **argv)
{
  int buffer_length;                      /* Number of bytes read in */
  char buffer[MAX_BUFF_SIZE];        /* Maximum size we will read in */
  char src_addr[INET_ADDRSTRLEN];         /* Where the NFE came from */
  socklen_t len;
  int sockfd, port = DEFAULT_SERV_PORT;
  struct sockaddr_in servaddr, cliaddr;
  char c;

  while ((c = getopt(argc, argv, "hp:")) != -1)
  {
    switch (c)
      {
      case 'h':
        usage(argv[0]);
        return 0;

      case 'p':
	port = atoi(optarg);
        break;

      case '?':
        usage(argv[0]);
	return 0;
      }
  }

  printf("Bare bones netflow collector %s\n", VERSION);

  printf("Written by Martin van den Nieuwelaar - martin@gadgets.co.nz\n");
  printf("Note: All timestamps are displayed in local time, not UTC!\n");

  sockfd = socket(AF_INET, SOCK_DGRAM, 0);

  bzero(&servaddr, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port = htons(port);

  printf("Binding to port %d...\n", port);

  if (bind(sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) == -1)
  {
    printf("Failed to bind to socket (port already in use?).  Exiting...\n");

    exit(1);
  }

  printf("Listening...\n");

  for ( ; ; )
  {
    len = sizeof(cliaddr);
    buffer_length = recvfrom(sockfd, buffer, MAX_BUFF_SIZE, 0,
			     (struct sockaddr *) &cliaddr, &len);

    inet_ntop(AF_INET, &cliaddr.sin_addr, src_addr, INET_ADDRSTRLEN);
    
    printf("Traffic received from IP address %s\n", src_addr);
    
    display_nfe(buffer, buffer_length);
  }
}
